#!/usr/bin/perl -w
# pconfig by Aaron D. Marasco (Aaron@Marascos.net) - No warranties expressed nor implied.
#  PLEASE send me your comments! http://www.fireparse.com/
#
# Please see the included README file for proper setup.
#
# 1.0 ADM -  9 May 2001 - I've been meaning to do this. Specifically written with fireparse in mind.
# 1.1 ADM - 21 May 2003 - Added output file option.
#
# Known Bugs/Concerns:
#   - None
#
# Future Enhancements:
#   - None I can think of!
#
# No, pconfig is NOT pconfig compliant! :(
##PCONFIG BEGIN NORMAL N/A
##PCONFIG ANNOUNCE #############################################################################
##PCONFIG ANNOUNCE #   This file does not use pconfig, please read the README file for setup!  #
##PCONFIG ANNOUNCE #############################################################################
##PCONFIG END NORMAL
##PCONFIG STOP
use strict;
my @infile; # A huge buffer for the file. Easiest way to manipulate it.
my $pointer = 0; # What line we are working on
my $in_block = 0; my $block_name = ""; # Are we listening or not?
my $ignore_block = 0; # Does the user want to process this block?
my $finish_note = ""; # If we want to have a finishing note
my $debug;# = 1; # For really loud and obnoxious set to one ;)
# For each entry:
my $descrip = ""; # The description text.
my (@accepts, @rejects); # Stored RegExs.
my $t_flag = ""; # Type, can be S or C for single or constant ('c' for no quotes). I to ignore the next attempt
my $prefix = ""; my $postfix = ""; # Text that can be ensured is before/after variable.
my $suf='pl'; # Default is to act as v1.0 - .pc -> .pl

# First let's suck in the entire file into $infile.
if (not defined $ARGV[0]) {
  print "pconfig takes a single argument of the perl script\nthat it should configure for you.\n";
  exit;
}
if ($ARGV[0] !~ /\.pc$/) {
  print "The perl script that is to be configured for you must end in '.pc'.\nThe output result is the same filename with a '.pl' extension.\n";
  exit;
}
{
  local $/ = undef; # Suck it all in in one shot.
  open (INFILEN, "< $ARGV[0]") or die "Couldn't open file: $ARGV[0]\nReason: $!";
  my $infile_temp = <INFILEN>;  # Get the whole thing
  close(INFILEN);
  @infile = split ("\n", $infile_temp);
}

POINTER: for ($pointer = 0; $pointer <= $#infile; $pointer++) {
#  $infile[$pointer] .= "\n"; # This was stripped out in the split above.
  if ($in_block == 0) {
    # Looking for a BEGIN or a STOP
    if ($infile[$pointer] =~ /^##PCONFIG\s+(BEGIN|STOP)\s*(\w*)\s*(.*)$/i) {
      last POINTER if ($1 eq "STOP");
      # Must be a "begin"
      if ($2 eq "") {
        warn "W: Block began on line $pointer with no name - defaulting to MAIN.\n";
        $block_name = "MAIN";
      } else {
        $block_name = $2;
      }
      print "\nBeginning block named $block_name\n";
      $in_block = 1;
      print "  $3\n" if ($3 ne "");
      print "Do you want to modify settings within this block? [Y/n]";
      my $tempval = <STDIN>;
      if ($tempval =~ /n/i) { # Any n is good enough for me
        $ignore_block = 1;
      } else {
        $ignore_block = 0;
      }
    } # Begin or stop
    elsif ($infile[$pointer] =~ /^##PCONFIG\s+EXT\s+(.*)$/) {
      $suf = $1;
    } # EXT command
    elsif ($infile[$pointer] =~ /^##PCONFIG\s+(.*)/i) {
      warn "W: Ignoring out-of-block directive on line $pointer: '$infile[$pointer]'\n";
    }
  } else { # Within a PCONFIG block now.
    # We have a few options at this point. We can have:
    # 1. Plain comment - ignore;
    # 2. Double-hash description - append to description
    # 3. Double-hash PCONFIG directive - must handle appropriately
    # 4. The actual thing that is being configured.
    #  a. Double-quoted value
    #  b. Single-quoted value
    #  c. constant
    my $myline = $infile[$pointer];
    study $myline;
    # Option 1: Plain comment (or blank line)
    next if (($myline =~ /^# /) or ($myline eq ""));
    # Option 2: Double hash description
    if (($myline =~ /^## (.*)/) && ($ignore_block == 0)) {
      # Check to make sure there is not a possible error.
      warn "W: Invalid whitespace detected:\n$pointer:\t$myline" if ($myline =~ /^##\s+PCONFIG/i);
      # Let's append this to the description.
      $descrip .= $1."\n";
      next POINTER;
    } # Option 2
    # Option 3: directive
    if ($myline =~ /^##PCONFIG\s+(.*)$/i) {
      my @directives = split (" ", $1); # Can't do uc here b/c of the regexs
      # Now parse each possible directive
      if ((uc $directives[0]) eq "END") { # End of block
        # Don't care about the name
        print "I: Ending block named $block_name\n" if (defined $debug);
        warn "W: ENDing $block_name before inserting variable with this description: (start)\n$descrip(end)\n" if $descrip;
        warn "W: ENDing $block_name before variable with ACCEPT clause seen\n" if (@accepts);
        warn "W: ENDing $block_name before variable with REJECT clause seen\n" if (@rejects);
        warn "W: ENDing $block_name before variable with special type flag seen (type $t_flag)\n" unless ($t_flag eq "");
        $in_block = 0;
        $block_name = "";
        print "\n", $finish_note, "\n" if ($finish_note and ($ignore_block == 0));
        $finish_note = "";
        # Clear all the info:
        &clear_info;
        next POINTER;
      } # End of block
      if (((uc $directives[0]) eq "REJECT") && ($ignore_block == 0)) { # This is a REJECT RegEx
        # Let's store the RegEx
        print "I: Storing REJECT statement: $directives[1]\n" if (defined $debug);
        push (@rejects, $directives[1]);
        next POINTER;
      } # REJECT
      if (((uc $directives[0]) eq "ACCEPT") && ($ignore_block == 0)) { # This is a ACCEPT RegEx
        # Let's store the RegEx
        print "I: Storing ACCEPT statement: $directives[1]\n" if (defined $debug);
        push (@accepts, $directives[1]);
        next POINTER;
      } # ACCEPT
      if (((uc $directives[0]) eq "IGNORE") && ($ignore_block == 0)) { # We should ignore next variable
        print "I: Going to ignore next variable assignment\n" if (defined $debug);
        warn "W: Overwriting previous special type flag ($t_flag) with (I)\n" unless ($t_flag eq "");
        $t_flag = "I";
        next POINTER;
      } # IGNORE
      if (((uc $directives[0]) eq "SINGLE") && ($ignore_block == 0)) { # Next variable is single quotes
        print "I: Marking next variable assignment as single-quoted\n" if (defined $debug);
        warn "W: Overwriting previous special type flag ($t_flag) with (S)\n" unless ($t_flag eq "");
        $t_flag = "S";
        next POINTER;
      } # SINGLE
      if (((uc $directives[0]) eq "CONSTANT") && ($ignore_block == 0)) { # Next entry is a constant
        print "I: Marking next variable assignment as constant\n" if (defined $debug);
        warn "W: Overwriting previous special type flag ($t_flag) with (C)\n" unless ($t_flag eq "");
        $t_flag = "C";
        next POINTER;
      } # CONSTANT
      if (((uc $directives[0]) eq "CONSTANTNOQ") && ($ignore_block == 0)) { # Next entry is a constant
        print "I: Marking next variable assignment as constant (noq)\n" if (defined $debug);
        warn "W: Overwriting previous special type flag ($t_flag) with (c)\n" unless ($t_flag eq "");
        $t_flag = "c";
        next POINTER;
      } # CONSTANTNOQ      
      if (((uc $directives[0]) eq "NAKED") && ($ignore_block == 0)) { # Next is a naked code line
        print "I: Marking next plain line as NAKED\n" if (defined $debug);
        warn "W: Overwriting previous special type flag ($t_flag) with (N)\n" unless ($t_flag eq "");
        $t_flag = "N";
        next POINTER;
      } # NAKED      
      if (((uc $directives[0]) eq "FINISH_NOTE") && ($ignore_block == 0)) { # Finishing notes (per block)
        $myline =~ /^##PCONFIG\s+FINISH_NOTE\s+(.*)$/i;
        if (not defined $1) {
          warn "W: Invalid finish note specified in line $pointer.\n";
          next POINTER;
        }
        $finish_note .= $1."\n";
        next POINTER;
      } # FINISH_NOTE      
      if ((uc $directives[0]) eq "ANNOUNCE") { # 'Blind' print statement (must be in a block but ignores overrides)
        $myline =~ /^##PCONFIG\s+ANNOUNCE\s+(.*)$/i;
        if (not defined $1) {
          warn "W: Invalid announcemnt specified in line $pointer.\n";
          next POINTER;
        }
        print $1, "\n";
        next POINTER;
      } # FINISH_NOTE            
      if ((uc $directives[0]) eq "REQUIRE") { # The remainder of this block is required now      
        if ((uc $directives[1]) eq "ONCE") { # Only need to do it once
          $infile[$pointer] = ""; # Wipe this line
          print "I: Marking remainder of $block_name as required ONCE\n" if (defined $debug);
        } else {
          print "I: Marking remainder of $block_name as required\n" if (defined $debug);
        }
        print "The remainder of $block_name has been deemed too important to skip!\n";
        $ignore_block = 0;
        # Clear all the info (bad info may have been stored off)
        &clear_info;
        next POINTER;
      } # REQUIRE
      if (((uc $directives[0]) eq "PREFIX") && ($ignore_block == 0)) { # Special prefix applied
        $myline =~ /^##PCONFIG\s+PREFIX\s+(.*)$/i;
        if (not defined $1) {
          warn "W: Invalid prefix specified in line $pointer.\n";
          next POINTER;
        }
        print "I: Adding prefix: '$1'\n" if (defined $debug);
        $prefix = $1;
        next POINTER;
      } # PREFIX
      if (((uc $directives[0]) eq "POSTFIX") && ($ignore_block == 0)) { # Special postfix applied
        $myline =~ /^##PCONFIG\s+POSTFIX\s+(.*)$/i;
        if (not defined $1) {
          warn "W: Invalid postfix specified in line $pointer.\n";
          next POINTER;
        }
        print "I: Adding postfix: '$1'\n" if (defined $debug);
        $postfix = $1;
        next POINTER;
      } # POSTFIX
      warn "W: Ignoring unhandled/invalid directive on line $pointer:\n\t@directives\n" unless ($ignore_block == 1);
      next POINTER;
    } # Option 3

    # All that's left is option #4. Let's get the variable name and current value.
    next POINTER if ($ignore_block == 1);
    if ($t_flag eq "I") { # Ignore this one
      print "I: Ignoring:\n\t$myline\n" if (defined $debug);
      # Clear all the stored info:
      &clear_info;
      next POINTER;
    };
    # All that is left now is single, double, constant
    my ($var_name, $prev_val);
INVALID:
    if ($t_flag eq "") {
      $myline =~ /^\s*(?:my)*\s*\$(\w+)\s*=\s*\"(.*)\"\s*;\s*$/i;
      if ((not defined $1) or (not defined $2)) {
        warn  "W: Improperly formatted variable string (double quoted):\n\t$infile[$pointer]\n";
        # Clear all the stored info:
        &clear_info;
        next POINTER;
      }
      $var_name = $1;
      $prev_val = $2;
    } elsif ($t_flag eq "S") { # single quotes
      $myline =~ /^\s*(?:my)*\s*\$(\w+)\s*=\s*\'(.*)\'\s*;\s*$/i;
      if ((not defined $1) or (not defined $2)) {
        warn  "W: Improperly formatted variable string (single quoted):\n\t$infile[$pointer]\n";
        # Clear all the stored info:
        &clear_info;
        next POINTER;
      }
      $var_name = $1;
      $prev_val = $2;
    } elsif ($t_flag eq "c") { # constant (noq)
      $myline =~ /^\s*use\s+constant\s+(\w+)\s*=>\s*(.*)\s*;\s*$/i;
      if ((not defined $1) or (not defined $2)) {
        warn  "W: Improperly formatted variable string (constant no quotes):\n\t$infile[$pointer]\n";
        # Clear all the stored info:
        &clear_info;
        next POINTER;
      }
      $var_name = $1;
      $prev_val = $2;
    } elsif ($t_flag eq "C") { # constant
      $myline =~ /^\s*use\s+constant\s+(\w+)\s*=>\s*'(.*)'\s*;\s*$/i;
      if ((not defined $1) or (not defined $2)) {
        warn  "W: Improperly formatted variable string (constant):\n\t$infile[$pointer]\n";
        # Clear all the stored info:
        &clear_info;
        next POINTER;
      }
      $var_name = $1;
      $prev_val = $2;
    } else {
      $myline =~ /^\Q$prefix\E(.*)\Q$postfix\E$/;
      if (not defined $1) {
        warn  "W: Improperly formatted naked string:\n\t$infile[$pointer]\n";
        exit;
        # Clear all the stored info:
        &clear_info;
        next POINTER;        
      }
      $prev_val = $1;
    }
    
    print '#'x50, "\n";
    print "\n", $descrip, "\n" if $descrip;
    print "Current value is: $prev_val\n Change it? [y/N] ";
    my $tempval = <STDIN>;
    if ($tempval !~ /y/i) { # Any y is good enough for me
      # Clear all the stored info:
      &clear_info;
      next POINTER;
    }
    print "       New value: ";
    $tempval = <STDIN>;
    chomp ($tempval);
    if ($prefix ne "") { $tempval = $prefix.$tempval unless ($tempval =~ /^\Q$prefix\E/) }
    if ($postfix ne "") { $tempval .= $postfix unless ($tempval =~ /\Q$postfix\E$/) }
    # Check accepts
    if (@accepts) {
      for (my $regex = 0; $regex <= $#accepts; $regex++) {
        print "I: Applying ACCEPT regex $accepts[$regex]\n" if (defined $debug);
        goto ACCEPTED if $tempval =~ /(?)$accepts[$regex]/i; # I know it is ugly but cannot figger another way.
      }
      print "Sorry, that choice seems to be invalid. Please try again.\n";
      goto INVALID;
    }
    # Check if blank (note that a regex checking for blanks will skip this!)
    if ($tempval =~ /^\s*$/) {
      print "Empty data is not acceptable here. Please try again.\n";
      goto INVALID;
    }
ACCEPTED:
    # Check rejects
    if (@rejects) {
      for (my $regex = 0; $regex <= $#rejects; $regex++) {
        print "I: Applying REJECT regex $rejects[$regex]\n" if (defined $debug);
        if ($tempval =~ /(?)$rejects[$regex]/i) {
          print "Sorry, that choice seems to be invalid. Please try again.\n";
          goto INVALID;
        }
      }
    }
    # Store the new value
    print "I: Replace this line:\n$infile[$pointer]\nwith this:\n" if (defined $debug);
    if ($myline =~ /^\s*my/) { # Need to put in a 'my'
      $infile[$pointer] = "my ";
    } else { # No 'my' needed
      $infile[$pointer] = "my ";
    }
    if ($t_flag eq "") {
      $infile[$pointer] .= "\$$var_name = \"$tempval\";"
    } elsif ($t_flag eq "S") {
      $infile[$pointer] .= "\$$var_name = '$tempval';"
    } elsif ($t_flag eq "c") { # Constant (noq)
      $infile[$pointer] = "use constant $var_name => $tempval;"    
    } elsif ($t_flag eq "C") { # Constant (will get rid of my if put in by accident)
      $infile[$pointer] = "use constant $var_name => '$tempval';"
    } else { # Naked line - leave it alone (prefix and postfix already handled.
      $infile[$pointer] = $tempval;
    }         
    print "$infile[$pointer]\n" if (defined $debug);
    # Clear all the stored info:
    &clear_info;
    next POINTER;
    # have some elsifs here for t_info
    warn "W: Wasn't expecting line $pointer right now:\n\t$infile[$pointer]\n";
  } # in_block
}

# Dump the new file
{
  for ($pointer = 0; $pointer <= $#infile; $pointer++) {
    $infile[$pointer].="\n"; # This was stripped way back when
  }  
  local $/ = undef; # Dump it all in in one shot.
  my $outfile = $ARGV[0];
  $outfile =~ s/pc$/$suf/;
  open (OUTFILEN, "> $outfile") or die "Couldn't create file: $outfile\nReason: $!";
  print OUTFILEN @infile;
  close(OUTFILEN);
}

print "pconfig Done\n";

sub clear_info {
  $descrip = "";
  undef @accepts;
  undef @rejects;
  $t_flag = "";
  $prefix = "";
  $postfix = "";
}
