#
# Another example plugin for ADMLogger for dhcpd admins
#
##PCONFIG BEGIN NORMAL N/A
##PCONFIG ANNOUNCE #############################################################################
##PCONFIG ANNOUNCE #   This file does not use pconfig, please read the README file for setup!  #
##PCONFIG ANNOUNCE #############################################################################
##PCONFIG END NORMAL
##PCONFIG STOP

package dhcpd;
use strict;
use Data::Dumper;

sub new {
    my ($logh, $mailh, $neth) = @_;
    my @regreq = ('dhcpd: DHCP');
    my $t = $logh -> register(\@regreq, 'dhcpd');
    print STDERR "DHCPD REGISTRATION ERROR!!! $t\n" unless ($t eq '');
    # Now we will create a hash listing all 'OK' MAC addresses.
    open(OKFILE, "< /etc/dhcpd.conf") or die "Couldn't open DHCP file for reading: $!\n";
    my %OKMACS;
    while (<OKFILE>) {
      next if /^\s*#/;
      next unless /hardware\s+ethernet\s+([A-Fa-f0-9:]{17})/; # MAC always 17 hex
      my $mac = lc $1;
      $OKMACS{$mac} = '1';
    }
    close OKFILE;
    return bless {
            'lh' => $logh,
            'mh' => $mailh,
            'nh' => $neth,
            'okmacs' => \%OKMACS
            }, 'dhcpd';
}

sub ready {
    my ($pkg) = @_;
    my $lh = $pkg->{'lh'};
    my $mh = $pkg->{'mh'};
    my $nh = $pkg->{'nh'};
    my %OKMACS = %{$pkg->{'okmacs'}};
    $mh -> setHTML(0); # We don't have HTML option
    my $date;
    chomp($date = `/bin/date +%m-%d-%Y`);
    my $host = `hostname`;
    chomp $host;
    my $mail_subject = "DHCP Report for -DATE- ($host)";
    if ($mail_subject =~ /(.*)-DATE-(.*)/) { $mail_subject = "$1$date$2"; };
    $mh->subject($mail_subject);
    my (%BADACKS, %BADDISCS, %BADREQS, %BADOFFERS);
    MYLINES:
    foreach (@{$lh -> retrieve_condensed('dhcpd')}) {
        my $count = $$_[0];
        my $str = lc $$_[2];
        $str =~ /([a-f0-9:]{17})/;
        # Example problem line:
        # hostname dhcpd: dhcpinform from 192.168.1.154 via eth8: not authoritative for subnet 192.168.1.0
        next MYLINES unless defined $1;
        next MYLINES if defined $OKMACS{$1};
        # ACK/DISCOVER
        #Apr 20 08:10:16 hostname dhcpd: DHCPACK on 192.168.n.n to 00:11:22:33:44:55 via eth8
        #Apr 21 15:10:17 hostname dhcpd: DHCPDISCOVER from 00:11:22:33:44:55 via eth8
        # Sometimes after the MAC there is a computer name
        if ($str =~ /DHCP(ACK|DISCOVER).*?([0-9A-Fa-f:]+)(?:\s*\(.*\))*\s+via\s+(.*)$/i) {
            if ($1 eq 'ack') {
                if (not exists $BADACKS{$2}) {
                    ${$BADACKS{$2}}[0] = 0;
                }
                ${$BADACKS{$2}}[0] += $count;
                ${$BADACKS{$2}}[1] = $3; # I know, a bug, oh well, last one wins
            } elsif ($1 eq 'discover') {
                if (not exists $BADDISCS{$2}) {
                    ${$BADDISCS{$2}}[0] = 0;
                }
                ${$BADDISCS{$2}}[0] += $count;
                ${$BADDISCS{$2}}[1] = $3;
            }
        } # ACK/DISCOVER
        #Apr 21 15:10:17 hostname dhcpd: DHCPOFFER on 192.168.n.n to 00:11:22:33:44:55 via eth8
        elsif ($str =~ /OFFER.*to\s+([^\s]+)(?:\s*\(.*\))*\s+via\s+(.*)$/i) {
            if (not exists $BADOFFERS{$1}) {
                ${$BADOFFERS{$1}}[0] = 0;
            }
            ${$BADOFFERS{$1}}[0] += $count;
            ${$BADOFFERS{$1}}[1] = $2;
        } # OFFER
        #Apr 20 08:10:16 hostname dhcpd: DHCPREQUEST for 192.168.n.n from 00:11:22:33:44:55 via eth8
        #Apr 21 15:10:17 hostname dhcpd: DHCPREQUEST for 192.168.n.n (192.168.n.n) from 00:11:22:33:44:55 via eth8
        elsif ($str =~ /REQUEST.*from\s+([^\s]+)(?:\s*\(.*\))*\s+via\s+(.*)$/i) {
            if (not exists $BADREQS{$1}) {
                ${$BADREQS{$1}}[0] = 0;
            }
            ${$BADREQS{$1}}[0] += $count;
            ${$BADREQS{$1}}[1] = $2;
        } # REQUEST
        elsif ($str =~ /NAK/i) {
            # Do nothing for NAK.
        } # NAK
        #else {
        #    print STDERR "Unhandled: $str\n";
        #} # Unhandled
    } # my lines
    foreach (sort keys %BADACKS) {
        $mh->append(${$BADACKS{$_}}[0]." unrecognized ACK(s) for $_ from ".${$BADACKS{$_}}[1]."\n");
    }
    foreach (sort keys %BADDISCS) {
        $mh->append(${$BADDISCS{$_}}[0]." unrecognized DISCOVER(s) for $_ from ".${$BADDISCS{$_}}[1]."\n");
    }
    foreach (sort keys %BADOFFERS) {
        $mh->append(${$BADOFFERS{$_}}[0]." unrecognized OFFER(s) for $_ from ".${$BADOFFERS{$_}}[1]."\n");
    }
    foreach (sort keys %BADREQS) {
        $mh->append(${$BADREQS{$_}}[0]." unrecognized REQ(s) for $_ from ".${$BADREQS{$_}}[1]."\n");
    }
    if ($mh -> isSendable()) {
        my $t = $mh->sendit();
        print STDERR "dhcp: Message sending failed - $t\n" if ($t ne '');
    }
} # sub ready

1;

