#
# Another example plugin for ADMLogger for pure-ftp admins
# Hopefully compared to fireparse this is easier to follow!
#
##PCONFIG BEGIN NORMAL N/A
##PCONFIG ANNOUNCE #############################################################################
##PCONFIG ANNOUNCE #   This file does not use pconfig, please read the README file for setup!  #
##PCONFIG ANNOUNCE #############################################################################
##PCONFIG END NORMAL
##PCONFIG STOP

package pureftp;
use strict;
#use Data::Dumper;

sub new {
    my ($logh, $mailh, $neth) = @_;
    # Every plugin must have a 'new' constructor that gets fed the ADMLogger handles that have already
    #  been opened and are ready. It is the responsibility of the plugin to register themselves under
    #  under the same name as their file.
    # You must always request an array even if it is one regex you want to match.
    my @regreq = ('pure-ftpd');
    my $t = $logh -> register(\@regreq, 'pureftp');
    print STDERR "PUREFTP REGISTRATION ERROR!!! $t\n" unless ($t eq '');
    # Now that we are registered, we just sit back and wait. We will later be called with ready().
    return bless {
            'lh' => $logh,
            'mh' => $mailh,
            'nh' => $neth,
            }, 'pureftp';
}

# This is the only other method that a plugin will have called. It is called after all data is ready
# for you. The plugin must get its own data from the ADMLogHandler since the caller doesn't know if
# you want to use 'retrieve' or 'retrieve_condensed'.
sub ready {
    my ($pkg) = @_;
    my $lh = $pkg->{'lh'};
    my $mh = $pkg->{'mh'};
    my $nh = $pkg->{'nh'};
    $mh -> setHTML(0); # We don't have HTML option
    my $date;
    chomp($date = `/bin/date +%m-%d-%Y`);
    my $host = `hostname`;
    chomp $host;
    my $mail_subject = "FTP Report for -DATE- ($host)";

    # Set up the displays
    my $NewMsg_TOP = '
                                  FTP REPORT
User                         Action    Thru           Filename
--------------------------------------------------------------------------------
';
    my $NewMsg_BODY = '
@<<<<<<<<<<<<<<<<<<<<<<<<<<< @<<<<<<<< @<<<<<<<<<<<<< @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
          @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
';

    {
        if ($mail_subject =~ /(.*)-DATE-(.*)/) { $mail_subject = "$1$date$2"; };
        $mh->subject($mail_subject);
        # This simple example ignores HTML
        $mh->append($NewMsg_TOP);
    }
    # Since we put in a header, but don't want that to be the only thing we send...
    $mh->clearflag();
    # This array will get all log lines we care about.
    my $filteredlines = $lh->retrieve('pureftp');
    if ($filteredlines eq '') {
        # Nothing to report!
        # May want to optionally report nothing some time in the future.
        return 1;
    }
    # Convert to array
    my @filterlines = split /\n/, $filteredlines;
    foreach (@filterlines) {
        my $line = $_;
        if ($line =~ /\((\w+?@[^)]*)\)\s+\[NOTICE\]\s+(.*)\s+([updown]+)loaded.*?,\s+(\d+\.\d+.*)\)/) {
        # Apr 26 12:33:27 hostname pure-ftpd[pid]: (user@host) [NOTICE] ///path/to/file/filename downloaded  (21451 bytes, 12612.23KB/sec)
        # Apr 15 21:40:53 hostname pure-ftpd[pid]: (user@host) [NOTICE] ///path/to/file/filename uploaded  (21652584 bytes, 11.00KB/sec)
            my @stack; # You'll love this one. ;)
            push @stack, $1; # user
            my $fname = $2;
            my $dir = $3;
            push @stack, $4; # thru
            $fname =~ s|/+|/|g; # only want one slash thanx
            $fname =~ /(.*)\/(.*)$/;
            $mh->append(swrite($NewMsg_BODY, shift @stack, ($dir eq 'up')?'Upload':'Download', shift @stack, $2, "$1/$2"));
        } elsif ($line =~ /\([\w\?]+?@[^)]*\)\s+\[WARNING\]\s+Authentication failed for user.*?\[(.*?)\]/) {
        # Apr 23 21:59:23 hostname pure-ftpd[pid]: (?@host) [WARNING] Authentication failed for user [user]
        # May 10 13:50:07 hostname pure-ftpd[pid]: (?@ip.ip.ip.ip) [WARNING] Authentication failed for user [user]        
            $mh->append(swrite($NewMsg_BODY, $1, 'Password FAIL', ' ', ' ', 'Authentication failed'));
        } elsif ($line =~ /\[INFO\]/) {
        # Most INFOs will be stripped out
        # Apr 23 06:43:27 hostname pure-ftpd[pid]: (user@host) [INFO] Can't change directory to /path/to/file/: No such file or directory
        # I had a permissions problem on the server, I want to know that:
        # Apr 22 05:38:47 hostname pure-ftpd[pid]: (user@host) [INFO] Can't change directory to /path/to/file/: Permission denied
            if ($line =~ /\(([\w@]+)[^\/]+(\/.*\/):\s+(.*denied)$/) {
                $mh->append(swrite($NewMsg_BODY, $1, 'ERROR', ' ', $2, $3));
            }
        } elsif ($line =~ /\(([\w\?]+?)@[^)]*\) \[NOTICE\] File successfully renamed or moved: \[([^\]]*)\][^\[]*\[([^\]])/) {    
        # May 11 10:04:22 hostname pure-ftpd[pid]: (user@host) [NOTICE] File successfully renamed or moved: [old]->[new]            
          $mh->append(swrite($NewMsg_BODY, $1, 'Rename', ' ', $2, "$2 -> $3"));
        } elsif ($line =~ /\(([\w\?]+?@[\w\?]+?)\)\s+\[([EROWANIG]+)\]\s+(.*)$/) {
        # Catch-all ERROR/WARNING not otherwise handled
        # Apr 26 12:39:05 hostname pure-ftpd[pid]: (user@host) [ERROR] Can't open that file: Permission denied
        # Apr 26 14:21:25 hostname pure-ftpd: (?@?) [ERROR] Unable to start a standalone server: [Address already in use]
        # Apr 24 16:34:42 hostname pure-ftpd[pid]: (user@host) [ERROR] Error during write to data connection: Broken pipe
        # Apr 26 14:21:25 hostname pure-ftpd: (?@?) [WARNING] Unknown run-time option
            $mh->append(swrite($NewMsg_BODY, $1, $2, ' ', ' ', $3));
        } elsif ($line =~ /pure-ftpd s\w+ succeeded/) { # do nothing (startup/shutdown)
        } else {
            # Anything left, should report just in case.
            $line =~ /pure-ftp(?:\[\d+\])?:(.*)$/;
            $mh->append(swrite($NewMsg_BODY, 'UNKNOWN', ' ', ' ', ' ', $line));

        } # if/elsif
    } # filterlines
    # Send the e-mail, if we did anything useful
    if ($mh->checkflag() == 1) {
        my $t = $mh->sendit();
        die "Message sending failed - $t\n" if ($t ne '');
    }
    ######################

    sub swrite {
        my $format = shift;
        $^A = "";
        formline($format,@_);
        return $^A;
    }

} # sub ready

1;
